"""Domain User class."""

from dataclasses import dataclass
from typing import List

from arrow import Arrow


@dataclass
class User:
    """Domain User class."""

    created_at: Arrow
    id_str: str
    rest_id_str: str
    default_profile: bool
    default_profile_image: bool
    description: str
    favourites_count: int
    followers_count: int
    friends_count: int
    has_custom_timelines: bool
    listed_count: int
    location: str
    media_count: str
    name: str
    pinned_tweet_ids_str: List[str]
    profile_banner_url: str
    profile_image_url_https: str
    protected: bool
    screen_name: str
    statuses_count: int
    verified: bool



"""TweetOutput which saves tweet data in memory, after request can return all saved tweets."""
from typing import List

from .tweet_output import TweetOutput
from ..model import Tweet


class CollectorTweetOutput(TweetOutput):
    """TweetOutput which print first tweet from request batch."""

    _tweets: List[Tweet]

    def __init__(self):
        """Constructor of class, initialize tweets as empty list."""
        self._tweets = []

    def export_tweets(self, tweets: List[Tweet]):
        """Method print first tweet in console, when tweets list is empty there message will be printed."""
        self._tweets.extend(tweets)
        return

    def get_scrapped_tweets(self) -> List[Tweet]:
        """Method returns all collected tweets."""
        return self._tweets



"""TweetOutput which print all tweets from request batch."""

from typing import List

from .tweet_output import TweetOutput
from ..model.tweet import Tweet


class PrintTweetOutput(TweetOutput):
    """TweetOutput which print all tweets from request batch."""

    def export_tweets(self, tweets: List[Tweet]):
        """Print all tweets from list."""
        for it in tweets:
            print(it)
        return



"""UserOutput which saves users as JSON objects."""

from typing import List

from .user_output import UserOutput
from ..mapper.user_json_mapper import user_to_json
from ..model import User


class JsonLineFileUserOutput(UserOutput):
    """UserOutput which saves users as JSON objects.

    UserOutput which saves user data by request received users batch to text file.
    Every user is saved in one line, this is JSON object.
    """

    file_name: str

    def __init__(self, file_name: str):
        """Creates instance of JsonLineFileUserOutput."""
        self.file_name = file_name

    def export_users(self, users: List[User]):
        """Append new user JSON strings to file."""
        with open(self.file_name, 'a') as file:
            for user in users:
                file.write(f'{user_to_json(user)}\n')
        return



"""RefreshTokenException definition."""


class RefreshTokenException(Exception):
    """RefreshTokenException class."""

    def __init__(self, msg):
        """Error constructor."""
        super().__init__(msg)



"""Abstract class for get guest auth token."""
from abc import abstractmethod

from ..http_request.web_client import WebClient


class AuthTokenProvider:
    """Abstract class for get guest auth token."""

    @abstractmethod
    def get_new_token(self) -> str:
        """Method returns new token."""


class AuthTokenProviderFactory:
    """Provider of AuthTokenProvider."""

    @abstractmethod
    def create(self, web_client: WebClient) -> AuthTokenProvider:
        """Method to create AuthTokenProvider from web_client."""



"""Runner for get tweets by ids."""
import traceback
from dataclasses import dataclass
from typing import Optional, List

from .get_user_request_details_builder import get_user_details_request_details
from .get_users_context import GetUsersContext
from .get_users_result import GetUsersResult
from .get_users_task import GetUsersTask
from .user_parser import parse_user
from ..auth import AuthTokenProviderFactory, SimpleAuthTokenProviderFactory
from ..http_request import WebClient
from stweet.http_request.requests.requests_web_client import RequestsWebClient
from ..model import User
from ..user_output import UserOutput


@dataclass
class _TweetByIdBaseInfo:
    id: str
    username: str
    tweet_content: str


class GetUsersRunner:
    """Runner class to process task to search tweets."""

    get_user_context: GetUsersContext
    get_user_task: GetUsersTask
    user_outputs: List[UserOutput]
    web_client: WebClient
    auth_token_provider_factory: AuthTokenProviderFactory

    def __init__(
            self,
            get_user_task: GetUsersTask,
            user_outputs: List[UserOutput],
            get_user_context: Optional[GetUsersContext] = None,
            web_client: WebClient = RequestsWebClient(),
            auth_token_provider_factory: AuthTokenProviderFactory = SimpleAuthTokenProviderFactory()
    ):
        """Constructor to create object."""
        self.get_user_context = GetUsersContext() if get_user_context is None else get_user_context
        self.get_user_task = get_user_task
        self.user_outputs = user_outputs
        self.web_client = web_client
        self.auth_token_provider_factory = auth_token_provider_factory
        return

    def run(self) -> GetUsersResult:
        """Main search_runner method."""
        self._prepare_token()
        for username in self.get_user_task.usernames:
            self._try_get_user(username)
        return GetUsersResult(self.get_user_context.scrapped_count)

    def _try_get_user(self, username: str):
        try:
            request_details = get_user_details_request_details(username, self.get_user_context)
            user_request_response = self.web_client.run_request(request_details)
            full_user = parse_user(user_request_response.text)
            self.get_user_context.add_one_scrapped_user()
            self._process_user_to_output(full_user)
        except Exception as e:
            print(f'error during scrap user -- {username}', e)
            traceback.print_exc()

    def _refresh_token(self):
        token_provider = self.auth_token_provider_factory.create(self.web_client)
        self.get_user_context.guest_auth_token = token_provider.get_new_token()
        return

    def _prepare_token(self):
        if self.get_user_context.guest_auth_token is None:
            self._refresh_token()
        return

    def _process_user_to_output(self, user: User):
        for user_output in self.user_outputs:
            user_output.export_users([user])



from .collector_user_output import CollectorUserOutput
from .csv_user_output import CsvUserOutput
from .json_line_file_user_output import JsonLineFileUserOutput
from .print_every_n_user_output import PrintEveryNUserOutput
from .print_user_output import PrintUserOutput
from .user_output import UserOutput



"""Domain Tweet class."""

from dataclasses import dataclass
from typing import List

from arrow import Arrow


@dataclass
class Tweet:
    """Domain Tweet class."""

    created_at: Arrow
    id_str: str
    conversation_id_str: str
    full_text: str
    lang: str
    favorited: bool
    retweeted: bool
    retweet_count: int
    favorite_count: int
    reply_count: int
    quote_count: int
    quoted_status_id_str: str
    quoted_status_short_url: str
    quoted_status_expand_url: str
    user_id_str: str
    user_name: str
    user_full_name: str
    user_verified: bool
    in_reply_to_status_id_str: str
    in_reply_to_user_id_str: str
    hashtags: List[str]
    mentions: List[str]
    urls: List[str]

    # def to_json_string(self) -> str:
    #     """Method to prepare json of tweet. Used in JSON serialization."""
    #     return json.dumps(self, default=lambda o: str(o) if isinstance(o, Arrow) else o.__dict__)
    #
    # def to_flat_dict(self):
    #     """Method to prepare flat dict of tweet. Used in CSV serialization."""
    #     dictionary = dict(self.__dict__)
    #     dictionary['hashtags'] = _simple_string_list_to_string(dictionary['hashtags'])
    #     dictionary['mentions'] = _simple_string_list_to_string(dictionary['mentions'])
    #     dictionary['urls'] = _simple_string_list_to_string(dictionary['urls'])
    #     return dictionary
    #
    # @staticmethod
    # def create_tweet_from_dict(dictionary: Dict[str, any]):
    #     """Method to create Tweet from dictionary."""
    #     return Tweet(
    #         arrow.get(dictionary['created_at']),
    #         str(dictionary['id_str']),
    #         str(dictionary['conversation_id_str']),
    #         dictionary['full_text'],
    #         dictionary['lang'],
    #         dictionary['favorited'],
    #         dictionary['retweeted'],
    #         dictionary['retweet_count'],
    #         dictionary['favorite_count'],
    #         dictionary['reply_count'],
    #         dictionary['quote_count'],
    #         dictionary['quoted_status_id_str'],
    #         dictionary['quoted_status_short_url'],
    #         dictionary['quoted_status_expand_url'],
    #         str(dictionary['user_id_str']),
    #         dictionary['user_name'],
    #         dictionary['user_full_name'],
    #         dictionary['user_verified'],
    #         str(dictionary['in_reply_to_status_id_str']),
    #         str(dictionary['in_reply_to_user_id_str']),
    #         dictionary['hashtags'],
    #         dictionary['mentions'],
    #         dictionary['urls']
    #     )
    #
    # @staticmethod
    # def create_tweet_from_flat_dict(dictionary: Dict[str, any]):
    #     """Method to create Tweet from flat dictionary."""
    #     dictionary['hashtags'] = _string_to_simple_string_list(dictionary['hashtags'])
    #     dictionary['mentions'] = _string_to_simple_string_list(dictionary['mentions'])
    #     dictionary['urls'] = _string_to_simple_string_list(dictionary['urls'])
    #     return Tweet.create_tweet_from_dict(dictionary)



"""Web client abstract class."""
from abc import abstractmethod

from .request_details import RequestDetails
from .request_response import RequestResponse


class WebClient:
    """Web client abstract class."""

    @abstractmethod
    def run_request(self, params: RequestDetails) -> RequestResponse:
        """Abstract method to run request."""



"""Util to process access token to Twitter api."""

import json
from json import JSONDecodeError

from .auth_token_provider import AuthTokenProvider, AuthTokenProviderFactory
from ..exceptions import RefreshTokenException
from ..http_request import HttpMethod
from ..http_request import RequestDetails, WebClient

_retries = 5
_timeout = 20
_url = 'https://api.twitter.com/1.1/guest/activate.json'
_auth_token = 'Bearer AAAAAAAAAAAAAAAAAAAAANRILgAAAAAAnNwIzUejRCOuH5E6I8xnZz4puTs%3D1Zv7ttfk8LF81' \
              'IUq16cHjhLTvJu4FA33AGWWjCpTnA'


class SimpleAuthTokenProvider(AuthTokenProvider):
    """Class to manage Twitter token api."""

    web_client: WebClient

    def __init__(self, web_client: WebClient):
        """Constructor of TokenRequest."""
        self.web_client = web_client
        return

    def _request_for_response_body(self):
        """Method from Twint."""
        token_request_details = RequestDetails(HttpMethod.POST, _url, {'Authorization': _auth_token}, dict(), _timeout)
        token_response = self.web_client.run_request(token_request_details)
        if token_response.is_success():
            return token_response.text
        else:
            raise RefreshTokenException('Error during request for token')

    def get_new_token(self) -> str:
        """Method to get refreshed token. In case of error raise RefreshTokenException."""
        try:
            token_html = self._request_for_response_body()
            return json.loads(token_html)['guest_token']
        except JSONDecodeError:
            raise RefreshTokenException('Error during request for token')


class SimpleAuthTokenProviderFactory(AuthTokenProviderFactory):
    """Provider of SimpleAuthTokenProviderFactory."""

    def create(self, web_client: WebClient) -> AuthTokenProvider:
        """Method to create SimpleAuthTokenProvider from web_client."""
        return SimpleAuthTokenProvider(web_client)



"""Generic class to process user output."""

from abc import ABC, abstractmethod
from typing import List

from ..model import User


class UserOutput(ABC):
    """Generic class to process user output."""

    @abstractmethod
    def export_users(self, tweets: List[User]):
        """Method to process all users from request batch."""



"""Domain SearchTweetsTask class."""

from dataclasses import dataclass
from typing import Optional

from arrow import Arrow

from .replies_filter import RepliesFilter
from ..model.language import Language


def _format_date(arrow_time: Arrow) -> int:
    return arrow_time.int_timestamp


@dataclass(frozen=True)
class SearchTweetsTask:
    """Domain SearchTweetsTask class."""

    all_words: Optional[str]
    exact_words: Optional[str]
    any_word: Optional[str]
    from_username: Optional[str]
    to_username: Optional[str]
    since: Optional[Arrow]
    until: Optional[Arrow]
    language: Optional[Language]
    tweets_limit: Optional[int]
    replies_filter: Optional[RepliesFilter]

    def __init__(
            self,
            all_words: Optional[str] = None,
            exact_words: Optional[str] = None,
            any_word: Optional[str] = None,
            from_username: Optional[str] = None,
            to_username: Optional[str] = None,
            since: Optional[Arrow] = None,
            until: Optional[Arrow] = None,
            language: Optional[Language] = None,
            tweets_limit: Optional[int] = None,
            replies_filter: Optional[RepliesFilter] = None
    ):
        """Class constructor."""
        object.__setattr__(self, 'all_words', all_words)
        object.__setattr__(self, 'exact_words', exact_words)
        object.__setattr__(self, 'any_word', any_word)
        object.__setattr__(self, 'from_username', from_username)
        object.__setattr__(self, 'to_username', to_username)
        object.__setattr__(self, 'since', since)
        object.__setattr__(self, 'until', until)
        object.__setattr__(self, 'language', language)
        object.__setattr__(self, 'tweets_limit', tweets_limit)
        object.__setattr__(self, 'replies_filter', replies_filter)
        return

    def get_full_search_query(self) -> str:
        """Method to return full search query."""
        query = ''
        if self.all_words is not None:
            query += self.all_words
        if self.exact_words is not None:
            query += f' "{self.exact_words}"'
        if self.any_word is not None:
            query += f' ({" OR ".join(self.any_word.split(" "))})'
        if self.language is not None:
            query += f' lang:{self.language.short_value}'
        if self.from_username:
            query += f' from:{self.from_username}'
        if self.since is not None:
            query += f" since:{_format_date(self.since)}"
        if self.until is not None:
            query += f" until:{_format_date(self.until)}"
        if self.to_username:
            query += f" to:{self.to_username}"
        if self.replies_filter is not None:
            if self.replies_filter == RepliesFilter.ONLY_REPLIES:
                query += " filter:replies"
            elif self.replies_filter == RepliesFilter.ONLY_ORIGINAL:
                query += " -filter:replies"
        return query



"""Domain GetUsersTask class."""
from dataclasses import dataclass
from typing import List


@dataclass(frozen=True)
class GetUsersTask:
    """Domain GetUsersTask class."""

    usernames: List[str]

    def __init__(
            self,
            usernames: List[str]
    ):
        """Class constructor."""
        object.__setattr__(self, 'usernames', usernames)
        return



"""User - dict mapper."""
from copy import copy
from typing import Dict

from arrow import get as arrow_get

from .util import simple_string_list_to_string, string_to_simple_string_list
from ..model import User


def user_to_flat_dict(self):
    """Method to prepare flat dict of tweet. Used in CSV serialization."""
    dictionary = dict(self.__dict__)
    dictionary['pinned_tweet_ids_str'] = simple_string_list_to_string(dictionary['pinned_tweet_ids_str'])
    return dictionary


def create_user_from_dict(dictionary: Dict[str, any]):
    """Method to create Tweet from dictionary."""
    return User(
        arrow_get(dictionary['created_at']),
        str(dictionary['id_str']),
        str(dictionary['rest_id_str']),
        dictionary['default_profile'],
        dictionary['default_profile_image'],
        dictionary['description'],
        dictionary['favourites_count'],
        dictionary['followers_count'],
        dictionary['friends_count'],
        dictionary['has_custom_timelines'],
        dictionary['listed_count'],
        dictionary['location'],
        dictionary['media_count'],
        dictionary['name'],
        dictionary['pinned_tweet_ids_str'],
        dictionary['profile_banner_url'],
        dictionary['profile_image_url_https'],
        dictionary['protected'],
        dictionary['screen_name'],
        dictionary['statuses_count'],
        dictionary['verified']
    )


def create_user_from_flat_dict(dictionary: Dict[str, any]) -> User:
    """Method to create Tweet from flat dictionary."""
    dictionary = copy(dictionary)
    dictionary['pinned_tweet_ids_str'] = string_to_simple_string_list(dictionary['pinned_tweet_ids_str'])
    return create_user_from_dict(dictionary)



"""Generic class to process tweet output."""

from abc import ABC, abstractmethod
from typing import List

from ..model.tweet import Tweet


class TweetOutput(ABC):
    """Generic class to process tweet output."""

    @abstractmethod
    def export_tweets(self, tweets: List[Tweet]):
        """Method to process all tweets from request batch."""



from .collector_tweet_output import CollectorTweetOutput
from .csv_tweet_output import CsvTweetOutput
from .json_line_file_tweet_output import JsonLineFileTweetOutput
from .print_every_n_tweet_output import PrintEveryNTweetOutput
from .print_first_in_request_tweet_output import PrintFirstInRequestTweetOutput
from .print_tweet_output import PrintTweetOutput
from .tweet_output import TweetOutput



"""Domain TweetsByIdsContext class."""

from dataclasses import dataclass
from typing import Optional


@dataclass
class TweetsByIdsContext:
    """Domain TweetsByIdsContext class."""

    guest_auth_token: Optional[str]
    all_download_tweets_count: int

    def __init__(
            self,
            guest_auth_token: Optional[str] = None,
            all_download_tweets: int = 0
    ):
        """Class constructor."""
        self.guest_auth_token = guest_auth_token
        self.all_download_tweets_count = all_download_tweets
        return

    def add_downloaded_tweets_count(self, new_tweets_count: int):
        """Add download tweet to context counter."""
        self.all_download_tweets_count += new_tweets_count



"""Runner to process task to search tweets."""

from typing import List, Optional

from .parse import BaseTweetParser
from .parse import TweetParser
from .request_details_builder import get_search_tweet_request_details
from .search_run_context import SearchRunContext
from .search_tweets_result import SearchTweetsResult
from .search_tweets_task import SearchTweetsTask
from ..auth import AuthTokenProviderFactory, SimpleAuthTokenProviderFactory
from ..exceptions.scrap_batch_bad_response import ScrapBatchBadResponse
from ..http_request.request_details import RequestDetails
from ..http_request.web_client import WebClient
from stweet.http_request.requests.requests_web_client import RequestsWebClient
from ..model.tweet import Tweet
from ..tweet_output.tweet_output import TweetOutput


class TweetSearchRunner:
    """Runner class to process task to search tweets."""

    search_run_context: SearchRunContext
    search_tweets_task: SearchTweetsTask
    tweet_outputs: List[TweetOutput]
    web_client: WebClient
    tweet_parser: TweetParser
    auth_token_provider_factory: AuthTokenProviderFactory

    def __init__(
            self,
            search_tweets_task: SearchTweetsTask,
            tweet_outputs: List[TweetOutput],
            search_run_context: Optional[SearchRunContext] = None,
            web_client: WebClient = RequestsWebClient(),
            tweet_parser: TweetParser = BaseTweetParser(),
            auth_token_provider_factory: AuthTokenProviderFactory = SimpleAuthTokenProviderFactory()
    ):
        """Constructor to create object."""
        self.search_run_context = SearchRunContext() if search_run_context is None else search_run_context
        self.search_tweets_task = search_tweets_task
        self.tweet_outputs = tweet_outputs
        self.web_client = web_client
        self.tweet_parser = tweet_parser
        self.auth_token_provider_factory = auth_token_provider_factory
        return

    def run(self) -> SearchTweetsResult:
        """Main search_runner method."""
        self._prepare_token()
        while not self._is_end_of_scrapping():
            self._execute_next_tweets_request()
        return SearchTweetsResult(self.search_run_context.all_download_tweets_count)

    def _is_end_of_scrapping(self) -> bool:
        ctx = self.search_run_context
        return \
            ctx.last_tweets_download_count == 0 \
            or (ctx.scroll_token is None) \
            or self.search_run_context.all_download_tweets_count == self.search_tweets_task.tweets_limit

    def _execute_next_tweets_request(self):
        request_params = self._get_next_request_details()
        response = self.web_client.run_request(request_params)
        if response.is_token_expired():
            self._refresh_token()
        elif response.is_success():
            parsed_tweets = self.tweet_parser.parse_tweets(response.text)
            self.search_run_context.scroll_token = self.tweet_parser.parse_cursor(response.text)
            self._process_new_scrapped_tweets(parsed_tweets)
        else:
            raise ScrapBatchBadResponse(response)
        return

    def _process_new_scrapped_tweets(self, parsed_tweets: List[Tweet]):
        self._process_new_tweets_to_output(parsed_tweets)
        self.search_run_context.last_tweets_download_count = len(parsed_tweets)
        self.search_run_context.add_downloaded_tweets_count(len(parsed_tweets))

    def _get_next_request_details(self) -> RequestDetails:
        return get_search_tweet_request_details(self.search_run_context, self.search_tweets_task)

    def _refresh_token(self):
        token_provider = self.auth_token_provider_factory.create(self.web_client)
        self.search_run_context.guest_auth_token = token_provider.get_new_token()
        return

    def _prepare_token(self):
        if self.search_run_context.guest_auth_token is None:
            self._refresh_token()
        return

    def _process_new_tweets_to_output(self, new_tweets: List[Tweet]):
        for tweet_output in self.tweet_outputs:
            tweet_output.export_tweets(new_tweets)
        return



"""Utils to parse data from web response."""

import json
from datetime import datetime
from typing import List, Optional, Dict

from arrow import Arrow
from dateutil import parser

from .tweet_parser import TweetParser
from ...model import Tweet

_Tweet_formats = {
    'datetime': '%Y-%m-%d %H:%M:%S %Z',
    'datestamp': '%Y-%m-%d',
    'timestamp': '%H:%M:%S'
}


def _default_string_value(string: Optional[str], default_value: str) -> str:
    return string if string is not None else default_value


def _starts_with_any_of(text: str, start_value: List[str]) -> bool:
    for it in start_value:
        if text.startswith(it):
            return True
    else:
        return False


def _timeline_entry_content_contains_tweet(entry: any) -> bool:
    return 'tweet' in entry['content']['item']['content']


class BaseTweetParser(TweetParser):
    """Utils class to parse data from web response."""

    def parse_tweets(self, response_text: str) -> List[Tweet]:
        """Method to extract tweets from web response."""
        # main method part from twint -- https://github.com/twintproject/twint
        response_json = json.loads(response_text)
        if len(response_json['globalObjects']['tweets']) == 0:
            return []
        feed = []
        for timeline_entry in response_json['timeline']['instructions'][0]['addEntries']['entries']:
            # this will handle the cases when the timeline entry is a tweet
            if _starts_with_any_of(timeline_entry['entryId'], ['sq-I-t-', 'tweet-']):
                if _timeline_entry_content_contains_tweet(timeline_entry):
                    _id = timeline_entry['content']['item']['content']['tweet']['id']
                    # skip the ads
                    if 'promotedMetadata' in timeline_entry['content']['item']['content']['tweet']:
                        continue
                elif 'tombstone' in timeline_entry['content']['item']['content'] and 'tweet' in \
                        timeline_entry['content']['item']['content']['tombstone']:
                    _id = timeline_entry['content']['item']['content']['tombstone']['tweet']['id']
                else:
                    _id = None
                if _id is None:
                    raise ValueError('Unable to find ID of tweet in timeline.')
                try:
                    temp_obj = response_json['globalObjects']['tweets'][_id]
                except KeyError:
                    print('encountered a deleted tweet with id {}'.format(_id))
                    continue
                temp_obj['user_data'] = response_json['globalObjects']['users'][temp_obj['user_id_str']]
                if 'retweeted_status_id_str' in temp_obj:
                    rt_id = temp_obj['retweeted_status_id_str']
                    _dt = response_json['globalObjects']['tweets'][rt_id]['created_at']
                    _dt = datetime.strptime(_dt, '%a %b %d %H:%M:%S %z %Y')
                    _dt = response_json(_dt)
                    _dt = str(_dt.strftime(_Tweet_formats['datetime']))
                    temp_obj['retweet_data'] = {
                        'user_rt_id': response_json['globalObjects']['tweets'][rt_id]['user_id_str'],
                        'user_rt': response_json['globalObjects']['tweets'][rt_id]['full_text'],
                        'retweet_id': rt_id,
                        'retweet_date': _dt,
                    }
                feed.append(BaseTweetParser._tweet_dict_to_tweet_object(temp_obj))
        return feed

    @staticmethod
    def _tweet_dict_to_tweet_object(tweet) -> Tweet:
        return Tweet(
            created_at=Arrow.fromdatetime(parser.parse(tweet['created_at'])),
            id_str=tweet['id_str'],
            conversation_id_str=tweet['conversation_id_str'],
            # there was the problem with pandas exporting because \r is old version of \n
            full_text=tweet['full_text'].replace('\r', '\n'),
            lang=tweet['lang'],
            favorited=tweet['favorited'],
            retweeted=tweet['retweeted'],
            retweet_count=tweet['retweet_count'],
            favorite_count=tweet['favorite_count'],
            reply_count=tweet['reply_count'],
            quote_count=tweet['quote_count'],
            quoted_status_id_str=BaseTweetParser._get_default_string_value_from_dict(
                tweet, 'quoted_status_id_str'),
            quoted_status_short_url=BaseTweetParser._get_default_string_value_from_dict(
                tweet, 'quoted_status_short_url'),
            quoted_status_expand_url=BaseTweetParser._get_default_string_value_from_dict(
                tweet, 'quoted_status_expand_url'),
            user_id_str=tweet['user_data']['id_str'],
            user_name=tweet['user_data']['screen_name'],
            user_full_name=tweet['user_data']['name'],
            user_verified=tweet['user_data']['verified'],
            in_reply_to_status_id_str=_default_string_value(tweet['in_reply_to_status_id_str'], ''),
            in_reply_to_user_id_str=_default_string_value(
                tweet['in_reply_to_user_id_str'], ''),
            hashtags=['#' + it['text'] for it in tweet['entities']['hashtags']],
            mentions=[it['screen_name'] for it in tweet['entities']['user_mentions']],
            urls=[it['url'] for it in tweet['entities']['urls']]
        )

    @staticmethod
    def _get_default_string_value_from_dict(tweet_dict: Dict[str, any], field: str, default_value: str = ''):
        return default_value if field not in tweet_dict else tweet_dict[field]

    @staticmethod
    def _parse_cursor_first_location(response_json: any) -> Optional[str]:
        try:
            return response_json['timeline']['instructions'][0]['addEntries']['entries'][-1]['content'][
                'operation']['cursor']['value']
        except KeyError:
            return None
        except IndexError:
            return None

    @staticmethod
    def _parse_cursor_second_location(response_json: any) -> Optional[str]:
        try:
            return response_json['timeline']['instructions'][-1]['replaceEntry']['entry']['content']['operation'][
                'cursor']['value']
        except KeyError:
            return None
        except IndexError:
            return None

    def parse_cursor(self, response_content: str) -> Optional[str]:
        """Method to extract next cursor to scrap request from web response."""
        response_json = json.loads(response_content)
        next_cursor = BaseTweetParser._parse_cursor_first_location(response_json)
        if next_cursor is None:
            next_cursor = BaseTweetParser._parse_cursor_second_location(response_json)
        return next_cursor



"""TweetOutput which saves tweets as JSON objects."""

from typing import List

from .tweet_output import TweetOutput
from ..mapper.tweet_json_mapper import tweet_to_json
from ..model.tweet import Tweet


class JsonLineFileTweetOutput(TweetOutput):
    """TweetOutput which saves tweets as JSON objects.

    TweetOutput which saves tweet data by request received tweets batch to text file.
    Every tweet is saved in one line, this is JSON object.
    """

    file_name: str

    def __init__(self, file_name: str):
        """Creates instance of JsonLineFileTweetOutput."""
        self.file_name = file_name

    def export_tweets(self, tweets: List[Tweet]):
        """Append new tweet JSON strings to file."""
        with open(self.file_name, 'a') as file:
            for tweet in tweets:
                file.write(f'{tweet_to_json(tweet)}\n')
        return



"""Class with response details."""

from dataclasses import dataclass
from typing import Optional


@dataclass
class RequestResponse:
    """Class with response details. Independent of web library implementation."""

    status_code: Optional[int]
    text: Optional[str]

    def is_token_expired(self) -> bool:
        """Method to check that is token_expired response status."""
        return self.status_code == 429

    def is_success(self) -> bool:
        """Method to check that response have success status."""
        return self.status_code is not None and self.status_code < 300



"""UserOutput which print all users from request batch."""

from typing import List

from .user_output import UserOutput
from ..model import User


class PrintUserOutput(UserOutput):
    """UserOutput which print all users from request batch."""

    def export_users(self, users: List[User]):
        """Print all users from list."""
        for it in users:
            print(it)
        return



"""Prepare params for request in TweetsByIdsRunner."""
import json

from .get_users_context import GetUsersContext
from ..http_request import HttpMethod
from ..http_request import RequestDetails

_graphql_token = 'esn6mjj-y68fNAj45x5IYA'
_request_url = f'https://api.twitter.com/graphql/{_graphql_token}/UserByScreenName'

_auth_token = 'Bearer AAAAAAAAAAAAAAAAAAAAANRILgAAAAAAnNwIzUejRCOuH5E6I8xnZz4p' \
              'uTs%3D1Zv7ttfk8LF81IUq16cHjhLTvJu4FA33AGWWjCpTnA'


def get_user_details_request_details(username: str, get_user_context: GetUsersContext) -> RequestDetails:
    """Prepare params for request in TweetsByIdsRunner."""
    return RequestDetails(
        http_method=HttpMethod.GET,
        url=_request_url,
        headers=dict({
            'x-guest-token': get_user_context.guest_auth_token,
            'Authorization': _auth_token
        }),
        params=dict({
            'variables': _get_variables_string(username)
        }),
        timeout=10
    )


def _get_variables_string(username: str) -> str:
    return json.dumps({'screen_name': username, 'withHighlightedLabel': True})



"""Domain SearchRunContext class."""

from dataclasses import dataclass
from typing import Optional

_init_scroll_token = '-1'


@dataclass
class SearchRunContext:
    """Domain SearchRunContext class."""

    scroll_token: Optional[str]
    guest_auth_token: Optional[str]
    last_tweets_download_count: int
    all_download_tweets_count: int

    def __init__(
            self,
            scroll_token: str = _init_scroll_token,
            guest_auth_token: Optional[str] = None,
            last_tweets_download_count: int = -1,
            all_download_tweets: int = 0
    ):
        """Class constructor."""
        self.scroll_token = scroll_token
        self.guest_auth_token = guest_auth_token
        self.last_tweets_download_count = last_tweets_download_count
        self.all_download_tweets_count = all_download_tweets
        return

    def add_downloaded_tweets_count(self, downloaded_tweets_count: int):
        """Method to update downloaded tweets count."""
        self.all_download_tweets_count += downloaded_tweets_count



import setuptools

with open("README.md", "r") as fh:
    LONG_DESCRIPTION = fh.read()

with open("requirements.txt", "r") as fh:
    REQUIRED = fh.readlines()

setuptools.setup(
    name="stweet",
    version="1.1.2",
    author="Marcin Wątroba",
    author_email="markowanga@gmail.com",
    description="Package to scrap tweets",
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    url="https://github.com/markowanga/stweet",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.7',
    install_requires=REQUIRED
)
